/*
 * yppasswdd.c	RPC Server routines.
 *
 * Copyright 1997 Thorsten Kukuk <kukuk@uni-paderborn.de>
 * Copyright 1994, 1995, 1996 Olaf Kirch, <okir@monad.swb.de>
 *
 * This program is covered by the GNU General Public License, version 2.
 * It is provided in the hope that it is useful. However, the author
 * disclaims ALL WARRANTIES, expressed or implied. See the GPL for details.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <termios.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <paths.h>
#include <getopt.h>
#include <syslog.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <rpc/rpc.h>
#include <rpc/pmap_clnt.h>
#include "yppasswd.h"
#include "version.h"


/* How often to retry locking the passwd file...
 */
#define MAX_RETRIES 5

static char *program_name = "";
static char *version = "yppasswdd " VERSION;

int	use_shadow = 0;
int	allow_chsh = 0,
	allow_chfn = 0;

#define xprt_addr(xprt)	(svc_getcaller(xprt)->sin_addr)
#define xprt_port(xprt)	ntohs(svc_getcaller(xprt)->sin_port)
void	yppasswdprog_1(struct svc_req *rqstp, SVCXPRT *transp);
void	reaper(int sig);

/*==============================================================*
 * RPC dispatch function
 *==============================================================*/
void
yppasswdprog_1(struct svc_req *rqstp, SVCXPRT *transp)
{
    union {
        yppasswd yppasswdproc_update_1_arg;
    } argument;
    char        *result;
    xdrproc_t   xdr_argument, xdr_result;
    char *(*local)();

    switch (rqstp->rq_proc) {
    case NULLPROC:
        svc_sendreply(transp, (xdrproc_t)xdr_void, (char *)NULL);
        return;

    case YPPASSWDPROC_UPDATE:
        xdr_argument = (xdrproc_t) xdr_yppasswd;
        xdr_result   = (xdrproc_t) xdr_int;
        local = (char *(*)()) yppasswdproc_pwupdate_1;
        break;

    default:
        svcerr_noproc(transp);
        return;
    }
    bzero((char *)&argument, sizeof(argument));
    if (!svc_getargs(transp, xdr_argument, (caddr_t)&argument)) {
        svcerr_decode(transp);
        return;
    }
    result = (*local)(&argument, rqstp);
    if (result != NULL
     && !svc_sendreply(transp, (xdrproc_t)xdr_result, result)) {
        svcerr_systemerr(transp);
    }
    if (!svc_freeargs(transp, xdr_argument, (caddr_t)&argument)) {
        (void)fprintf(stderr, "unable to free arguments\n");
        exit(1);
    }
}

static void
usage(FILE *fp, int n)
{
    fprintf (fp, "usage: %s [-h] [-v]\n", program_name );
    exit(n);
}

void
sig_child (int sig)
{
    wait(NULL);
}

/* Clean up if we quit the program. */
void
sig_quit (int sig)
{
  pmap_unset(YPPASSWDPROG, YPPASSWDVERS);
  exit(0);
}


void
install_sighandler (void)
{
  struct sigaction sa;

  sigaction(SIGPIPE, NULL, &sa);
  sa.sa_handler = SIG_IGN;
#if !defined(sun) || (defined(sun) && defined(__svr4__))
  sa.sa_flags |= SA_RESTART;
  /* The opposite to SA_ONESHOT, do  not  restore
     the  signal  action.  This provides behavior
     compatible with BSD signal semantics. */
#endif
  sigemptyset(&sa.sa_mask);
  sigaction(SIGPIPE, &sa, NULL);
  /* Clear up if child exists */
  sigaction(SIGCHLD, NULL, &sa);
#if !defined(sun) || (defined(sun) && defined(__svr4__))
  sa.sa_flags |= SA_RESTART;
#endif
  sa.sa_handler = sig_child;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGCHLD, &sa, NULL);
  /* If program quits, give ports free. */
  sigaction(SIGTERM, NULL, &sa);
#if !defined(sun) || (defined(sun) && defined(__svr4__))
  sa.sa_flags |= SA_RESTART;
#endif
  sa.sa_handler = sig_quit;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGTERM, &sa, NULL);

  sigaction(SIGINT, NULL, &sa);
#if !defined(sun) || (defined(sun) && defined(__svr4__))
  sa.sa_flags |= SA_RESTART;
#endif
  sa.sa_handler = sig_quit;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGINT, &sa, NULL);
}


int
main(int argc, char **argv)
{
  SVCXPRT *	transp;
  char *	sp;
  int		opterr;
  int		c;

  program_name = argv[0];
  if ((sp = strrchr(program_name, '/')) != NULL)
    program_name = ++sp;

  /* Parse the command line options and arguments. */
  opterr = 0;
  while ((c = getopt(argc, argv, "e:p:s:hv")) != EOF)
    switch (c)
      {
      case 'e':
	if (!strcmp(optarg, "chsh"))
	  allow_chsh = 1;
	else if (!strcmp(optarg, "chfn"))
	  allow_chfn = 1;
	else
	  usage(stderr, 1);
	break;
      case 'p':
	path_passwd = optarg;
	break;
      case 's':
	path_shadow = optarg;
	break;
      case 'h':
	usage(stdout, 0);
	break;
      case 'v':
	printf("%s\n", version);
	exit(0);
      case 0:
	break;
      case '?':
      default:
	usage(stderr, 1);
      }

    /* No more arguments allowed. */
    if (optind != argc)
        usage(stderr, 1);

    /* Create tmp file names */
    path_passwd_tmp = malloc (strlen (path_passwd) + 5);
    if (path_passwd_tmp == NULL)
      {
	fputs ("yppasswdd: out of memory\n", stderr);
	exit (-1);
      }
    sprintf (path_passwd_tmp, "%s.tmp", path_passwd);
    path_shadow_tmp = malloc (strlen (path_shadow) + 5);
    if (path_shadow_tmp == NULL)
      {
	fputs ("yppasswdd: out of memory\n", stderr);
	exit (-1);
      }
    sprintf (path_shadow_tmp, "%s.tmp", path_shadow);
    /* Create .OLD file names */
    path_passwd_old = malloc (strlen (path_passwd) + 5);
    if (path_passwd_old == NULL)
      {
	fputs ("yppasswdd: out of memory\n", stderr);
	exit (-1);
      }
    sprintf (path_passwd_old, "%s.OLD", path_passwd);
    path_shadow_old = malloc (strlen (path_shadow) + 5);
    if (path_shadow_old == NULL)
      {
	fputs ("yppasswdd: out of memory\n", stderr);
	exit (-1);
      }
    sprintf (path_shadow_old, "%s.OLD", path_shadow);

#ifndef RPC_SVC_FG
    /* We first fork off a child. */
    if ((c = fork()) > 0)
        exit(0);
    if (c < 0) {
        fprintf(stderr, "yppasswdd: cannot fork: %s\n", strerror(errno));
        exit(-1);
    }
    /* Now we remove ourselves from the foreground. */
    close(0);
    close(1);
    close(2);
#ifdef TIOCNOTTY
    if ((c = open("/dev/tty", O_RDWR)) >= 0) {
        ioctl(c, TIOCNOTTY, (char *) NULL);
        close(c);
    }
#else
    setsid();
#endif
#endif /* not RPC_SVC_FG */

    /* Initialize logging.
     */
    openlog ( "yppasswdd", LOG_PID, LOG_AUTH );

    /* Register a signal handler to reap children after they terminated
     */
    install_sighandler();

    /*
     * Create the RPC server
     */
    pmap_unset(YPPASSWDPROG, YPPASSWDVERS);

    transp = svcudp_create(RPC_ANYSOCK);
    if (transp == NULL) {
        fprintf(stderr, "cannot create udp service.\n");
        exit(1);
    }
    if (!svc_register(transp, YPPASSWDPROG, YPPASSWDVERS, yppasswdprog_1,
            IPPROTO_UDP)) {
        fprintf(stderr, "unable to register yppaswdd udp service.\n");
        exit(1);
    }

    /*
     * Run the server
     */
    svc_run();
    fprintf(stderr, "svc_run returned\n");

    return 1;
}
